/*****************************************************************************
 * BF70xRSAexample.c
 *****************************************************************************/

#include <sys/platform.h>
#include "RSAexample.h"
#include "pka_functions.h"
#include "math_utilities.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>

/** 
 * If you want to use command program arguments, then place them in the following string. 
 */
char __argv_string[] = "";


/*
 * Allocate buffer for p and q.  Since n = p * q, the bit length of p and q will be
 * half of the bit length of n
 */
section ("seg_l2_dmda_bw")
uint32_t num_p[MOD_SIZE_WORDS_BY2];
section ("seg_l2_dmda_bw")
uint32_t num_q[MOD_SIZE_WORDS_BY2];
section ("seg_l2_dmda_bw")
uint32_t num_dp[MOD_SIZE_WORDS_BY2];
section ("seg_l2_dmda_bw")
uint32_t num_dq[MOD_SIZE_WORDS_BY2];
section ("seg_l2_dmda_bw")
uint32_t num_n[MOD_SIZE_WORDS+4];
section ("seg_l2_dmda_bw")
uint32_t num_phi[MOD_SIZE_WORDS];
section ("seg_l2_dmda_bw")
uint32_t num_exp[MOD_SIZE_WORDS];
section ("seg_l2_dmda_bw")
uint32_t num_d[MOD_SIZE_WORDS];
section ("seg_l2_dmda_bw")
uint32_t num_p_min1[MOD_SIZE_WORDS_BY2];
section ("seg_l2_dmda_bw")
uint32_t num_q_min1[MOD_SIZE_WORDS_BY2];
section ("seg_l2_dmda_bw")
uint32_t num_q_inv[MOD_SIZE_WORDS_BY2];

section ("seg_l2_dmda_bw")
uint32_t msg_buffer[MOD_SIZE_WORDS];
section ("seg_l2_dmda_bw")
uint32_t ciphertext[MOD_SIZE_WORDS];
section ("seg_l2_dmda_bw")
uint32_t plaintext[MOD_SIZE_WORDS];
section ("seg_l2_dmda_bw")
uint32_t plaintext2[MOD_SIZE_WORDS];



bool genPrimeNum(uint32_t *num_buffer, uint32_t bitlength);
bool genRandNum(uint32_t *num_buffer, uint32_t bitlength);
bool genNumExp(uint32_t *num_buffer, uint32_t *phi_buffer, uint32_t bitlength);
bool genNumModInv(uint32_t *exp, uint32_t *phi, uint32_t szWds, uint32_t *d);

bool getModInv(uint32_t *num, uint32_t numSzWds, uint32_t *modulo, uint32_t moduloSzWds, uint32_t *modInvRes);
bool genKeyPair(keypair_t *kp, uint32_t szBits);

bool genMsg(uint32_t *msg, uint32_t *modulo, uint32_t maxSzBits);

bool rsaEncrypt(keypair_t *kp, uint32_t *msg_buffer, uint32_t *ct, uint32_t szWds);
bool rsaDecrypt(keypair_t *kp, uint32_t *ct, uint32_t *pt, uint32_t szWds, enum decmode mode);

section ("seg_l2_dmda_bw")
keypair_t kp;

void testDecryptedOutput(uint32_t *buf1,  uint32_t *buf2, uint32_t wds);



int main(int argc, char *argv[])
{

	uint32_t t;
	int i;

	kp.n = num_n;
	kp.pub_exp = num_exp;
	kp.priv_exp = num_d;
	kp.p = num_p;
	kp.q = num_q;
	kp.phi = num_phi;
	kp.p_min1 = num_p_min1;
	kp.q_min1 = num_q_min1;
	kp.dp = num_dp;
	kp.dq = num_dq;
	kp.qinv = num_q_inv;


	/*
	 * Generate the key pair.  Pick an exponent e that is  1 < e < φ(n).
	 * Then from e, calculate the modular inverse given that e and φ(n)
	 * are co-prime.
	 */
	genKeyPair(&kp, MOD_SIZE_BITS);


	/*
	 * Generate a "message" to encrypt
	 */
	genMsg(msg_buffer, num_n, MOD_SIZE_BITS);


	/*
	 * Encrypt
	 */
	rsaEncrypt(&kp, msg_buffer, ciphertext, MOD_SIZE_WORDS);

	/*
	 * Decrypt using pair (d, n)
	 */
	rsaDecrypt(&kp, ciphertext, plaintext, MOD_SIZE_WORDS, pair);
	testDecryptedOutput(msg_buffer, plaintext, MOD_SIZE_WORDS);

	/*
	 * Decrypt using 5-tuple (p, q, dP, dQ, qInv)
	 */
	rsaDecrypt(&kp, ciphertext, plaintext2, MOD_SIZE_WORDS, crt);
	testDecryptedOutput(msg_buffer, plaintext2, MOD_SIZE_WORDS);


	printf("...end\n");


	return 0;
}



/*
 *
 * genPrimeNum()
 *
 * Generate a large prime number.  This is done by generating a random number and test for primality.
 * This is a simple, non-efficient method and only used for demonstration purposes.
 *
 * Input
 * 	num_buffer - ptr to where the number is stored
 * 	bitlength  - size of prime in bits.
 *
 *
 */
bool genPrimeNum(uint32_t *num, uint32_t bitlength){

	int i;
	uint32_t numwords = (bitlength+31) / 32;           // round up to next 32 bits so we have enough words for bitlength
	bool res;

	do{
		// get a random number
		res = genRandNum(num, bitlength);

		if(res != true){
			printf("Error generating random number\n");
			return false;
		}
	}while(isPrime(num, numwords) == false);


	return true;
}





/*
 * Use TRNG to generate random number.
 *
 * TODO: take care of error conditions
 */
bool genRandNum(uint32_t *num_buffer, uint32_t bitlength){

	int i, j, k;
	uint32_t hibits, msw_idx;
	uint32_t *n = num_buffer;


#define MAX_SHUTDOWN_THRESHOLD 8
	*pREG_TRNG0_ALMCNT |= (MAX_SHUTDOWN_THRESHOLD << BITP_TRNG_ALMCNT_SHDNTHRESH);


	*pREG_TRNG0_FROEN = 0xFF;                 // make sure all 8 FRO's are on
	*pREG_TRNG0_CTL |= BITM_TRNG_CTL_TRNGEN;  // enable TRNG


	for(k = 0; k < bitlength; k += RAND_NUM_LEN_BITS){

		do {
			for(i=0; i<100; i++)                      // only try getting random #'s discrete # of times before quitting
				if((*pREG_TRNG0_STAT & 0xFF) == 0)
					for(j=0; j<1000; j++)             // delay loop
						;
				else
					break;

			// check for error
			if((*pREG_TRNG0_STAT & (BITM_TRNG_STAT_SHDNOVR  | \
					BITM_TRNG_STAT_STUCKOUT | \
					BITM_TRNG_STAT_NOISEFAIL| \
					BITM_TRNG_STAT_RUNFAIL  | \
					BITM_TRNG_STAT_LRUNFAIL | \
					BITM_TRNG_STAT_PKRFAIL  | \
					BITM_TRNG_STAT_MBITFAIL)) != 0){
#if defined(DEBUG) && DEBUG >=3
				printf("Error in TRNG: STAT: 0x%08x\n", *pREG_TRNG0_STAT);
				printf("Error in TRNG: ALMCNT.SHTDNTHRESH 0x%02x\n",
						(*pREG_TRNG0_ALMCNT & BITM_TRNG_ALMCNT_SHDNTHRESH) >> BITP_TRNG_ALMCNT_SHDNTHRESH);
				printf("Error in TRNG: ALMCNT.SHTDNCNT 0x%02x\n",
						(*pREG_TRNG0_ALMCNT & BITM_TRNG_ALMCNT_SHDNCNT) >> BITP_TRNG_ALMCNT_SHDNCNT);
#endif
				// FIXME : take care of errors
			}

			// check if random bits available
			if((*pREG_TRNG0_STAT & BITM_TRNG_STAT_RDY) != 0){
#if defined(DEBUG) && DEBUG >= 4
				printf("Rand # avail\n");
#endif
				if(((int32_t)bitlength - k - 0*32) > 0)
					num_buffer[0] = *pREG_TRNG0_OUTPUT0;
				if(((int32_t)bitlength - k - 1*32) > 0)
					num_buffer[1] = *pREG_TRNG0_OUTPUT1;
				if(((int32_t)bitlength - k - 2*32) > 0)
					num_buffer[2] = *pREG_TRNG0_OUTPUT2;
				if(((int32_t)bitlength - k - 3*32) > 0)
					num_buffer[3] = *pREG_TRNG0_OUTPUT3;

				num_buffer += (RAND_NUM_LEN_BITS/32);       // move buffer forward

				*pREG_TRNG0_INTACK |= BITM_TRNG_INTACK_RDY; // acknowledge reception of current data.  ready for new random bits

				break;                                      // move on to get next set of random bits
			}
		}while(1);

	}


	// clear off any MSB's that are not needed
	hibits = bitlength % 32;
	msw_idx = ((bitlength+31)/32) - 1;
	n[msw_idx] &= ((0xffffffff >> (32-hibits)));


	*pREG_TRNG0_CTL &= (~BITM_TRNG_CTL_TRNGEN);  // turn off TRNG


	return true;
}

/*
 * Generate Key Pair
 * Given n and phi, we pick the appropriate e and calculate the corresponding d
 */
bool genKeyPair(keypair_t *kp, uint32_t szBits){

	int i;
	uint32_t t, *num_tmp, *num_phi2;
	uint32_t num1=1;

	printf("genKeyPair(): Generating Key Pair\n");
	/*
	 * First thing we need is to generate some prime numbers
	 * Since |n| is szBits, p and q are half of szBits
	 */
	printf("genKeyPair(): Generating p\n");
	genPrimeNum(kp->p, szBits/2);

	printf("genKeyPair(): Generating q\n");
	genPrimeNum(kp->q, szBits/2);

	/*
	 * Compare and swap if necessary.  For form 2 where the private key is a 5-tuple,
	 * P needs to be greater than Q to be used with PKA.
	 */
	i = pka_cmp(kp->p, kp->q, MOD_SIZE_WORDS_BY2);
	if(i==0){
		printf("genKeyPair(): Error: p and q can't be the same!\n");
		exit(1);
	} else
		if(i == -1){
			uint32_t temp;
			printf("genKeyPair(): p<q -> need to swap in order to calculate 5-tuple private key\n");
			for(t=0; t<MOD_SIZE_WORDS_BY2; t++){
				temp = kp->p[t];
				kp->p[t] = kp->q[t];
				kp->q[t] = temp;
			}
		} else
			if(i == -2){
				printf("genKeyPair(): Error: error comparing p and q\n");
				exit(1);
			}

#ifdef DEBUG
	printf("p: \n\t");
	for(i=MOD_SIZE_WORDS_BY2-1; i>=0; i--) printf("%08x", kp->p[i]);  printf("\n");
	printf("q: \n\t");
	for(i=MOD_SIZE_WORDS_BY2-1; i>=0; i--) printf("%08x", kp->q[i]);  printf("\n");
#endif

	/*
	 * Calculate the modulus n = p*q
	 */
	pka_mult(kp->p, MOD_SIZE_WORDS_BY2, kp->q, MOD_SIZE_WORDS_BY2, kp->n);

#ifdef DEBUG
	printf("n = p * q: \n\t");
	for(i=MOD_SIZE_WORDS-1; i>=0; i--) printf("%08x", kp->n[i]);  printf("\n");
#endif

	/*
	 * Using the Euler's totient/phi function compute
	 * φ(n) = φ(p)φ(q) = (p − 1)(q − 1) = n - (p + q -1)
	 */
	pka_sub(kp->p, MOD_SIZE_WORDS_BY2, &num1, 1, kp->p_min1);    // p-1
	pka_sub(kp->q, MOD_SIZE_WORDS_BY2, &num1, 1, kp->q_min1);    // q-1

	pka_mult(kp->p_min1, MOD_SIZE_WORDS_BY2, kp->q_min1, MOD_SIZE_WORDS_BY2, kp->phi);

#ifdef DEBUG
	printf("phi(n) = (p-1) * (q-1): \n\t");
	for(i=MOD_SIZE_WORDS-1; i>=0; i--) printf("%08x", kp->phi[i]);  printf("\n");
#endif


	/*
	 * For verification, we can check that num_phi that we calculated by (p-1)(q-1)
	 * is equal to  n - (p + q -1)
	 */
#ifdef DEBUG
	printf("Info: Verifying Euler's phi: (p-1)(q-1) == n-(p+q-1) ?\n");
#endif
	num_tmp = malloc(MOD_SIZE_WORDS_BY2 * 4);
	num_phi2 = malloc(MOD_SIZE_WORDS * 4);
	if(num_tmp == NULL || num_phi2 == NULL){
		printf("Error: Couldn't allocate memory for num_phi validation.  skipping.\n");
	}
	else{

		pka_add(kp->p, MOD_SIZE_WORDS_BY2, num_q_min1, MOD_SIZE_WORDS_BY2, num_tmp);

		pka_sub(kp->n, MOD_SIZE_WORDS, num_tmp, MOD_SIZE_WORDS_BY2, num_phi2);

		if(pka_cmp(kp->phi, num_phi2, MOD_SIZE_WORDS) != 0)
			printf("Error: Calculations of Euler's phi do not match!\n");
		else
			printf("Info: Calculations of Euler's phi match!\n");

		free(num_tmp);
		free(num_phi2);
	}

	/* generate the exponent e such that 1 < e < φ(n) and GCD(e, φ(n))=1 */
	genNumExp(kp->pub_exp, kp->phi, szBits);

#ifdef DEBUG
	printf("e: (1 < e < phi(n) and GCD(e, phi(n))=1)\n\t");
	for(i=((szBits+31)/32)-1; i>=0; i--) printf("%08x", kp->pub_exp[i]);  printf("\n");
#endif


#if 0
	num_p[1] = 0x3;
	num_p[0] = 0xfe645e9d;

	num_q[1] = 0x1;
	num_q[0] = 0xf9fc49ab;

	num_phi[2] = 0x7;
	num_phi[1] = 0xe4c38fa0;
	num_phi[0] = 0x6b314f98;

	num_n[2] = 0x7;
	num_n[1] = 0xe4c38fa6;
	num_n[0] = 0x6391f7df;

	num_exp[2] = 0x0;
	num_exp[1] = 0x7f2faca1;
	num_exp[0] = 0xe033011f;

#endif

	printf("Calculating d (e*d = 1 mod (phi):\n");
	/* generate the exponent d by finding the inverse of e modulo φ(n) */
	genNumModInv(kp->pub_exp, kp->phi, (szBits+31)/32, kp->priv_exp);

#ifdef DEBUG
		printf("d:\n\t");
		for(i=((szBits+31))/32-1; i>=0; i--) printf("%08x", kp->priv_exp[i]); printf("\n");
#endif


	/*** compute the values for the alternative form of the private key ***/

	/* calculated dp = d mod (p-1) */
	if(MOD_SIZE_WORDS_BY2 > 1){
		pka_div(kp->priv_exp, MOD_SIZE_WORDS, kp->p_min1, MOD_SIZE_WORDS_BY2, NULL, kp->dp);
		pka_div(kp->priv_exp, MOD_SIZE_WORDS, kp->q_min1, MOD_SIZE_WORDS_BY2, NULL, kp->dq);
	}
	else{
		kp->dp[0] = aModB(kp->priv_exp, MOD_SIZE_WORDS, kp->p_min1[0]);
		kp->dq[0] = aModB(kp->priv_exp, MOD_SIZE_WORDS, kp->q_min1[0]);
	}

#ifdef DEBUG
	printf("dp: \n\t");
	for(i=MOD_SIZE_WORDS_BY2-1; i>=0; i--) printf("%08x", kp->dp[i]);  printf("\n");
	printf("dq: \n\t");
	for(i=MOD_SIZE_WORDS_BY2-1; i>=0; i--) printf("%08x", kp->dq[i]);  printf("\n");
	printf("calculating q^-1\n");
#endif

	/* calculate q^-1 mod p */
	genNumModInv(kp->q, kp->p, MOD_SIZE_WORDS_BY2, kp->qinv);

#ifdef DEBUG
	printf("q^-1: \n\t");
	for(i=MOD_SIZE_WORDS_BY2-1; i>=0; i--) printf("%08x", kp->qinv[i]);  printf("\n");
#endif

	return true;
}


/*
 * Generate the exponent for the public key
 * There are two requirements.
 *  1. 1 < exponent < num_phi
 *  2. GCD(exponent, num_phi) = 1 (co-prime)
 *
 */
bool genNumExp(uint32_t *num_buffer, uint32_t *phi_buffer, uint32_t bitlength){

	uint32_t *num1, numWds;
	uint32_t *gcdResult, *modInvRes, result;
	int i;

	numWds = (bitlength+31)/32;

	modInvRes = malloc((bitlength+7)/8);
	if(modInvRes == NULL){
		printf("Error: Can't allocate buffer for GCD result\n");
		exit(1);
	}

	num1 = malloc((bitlength+7)/8);
	if(num1 == NULL){
		printf("Error: Can't allocate buffer for NUM1\n");
		exit(1);
	}
	else{
		num1[0] = 1;
		for(i=1; i<numWds; i++)
			num1[i] = 0;
	}

	printf("genNumExp(): Generating e:\n");
	do{
		/* get a random number */
		if(genRandNum(num_buffer, bitlength) != true){
			printf("Error generating random number\n");
			return false;
		}

		/* make the random number odd since φ(n) = (p-1)(q-1) is even */
		if((num_buffer[0] & 1) == 0)
			num_buffer[0] |= 1;

		printf("\t"); for(i=numWds-1; i>=0; i--) printf("%08x", num_buffer[i]); printf("\n");

		if(pka_cmp(num1, num_buffer, numWds) != -1){
			printf("1 > e\n");
			//printf(".");
			continue;
		}

		if(pka_cmp(num_phi, num_buffer, numWds) != 1){
			printf("e > phi\n");
			//printf(".");
			continue;
		}

		result = pka_modinv(num_phi, numWds, num_buffer, numWds, modInvRes);
		if(result == 7){
			printf("GCD(e, phi) != 1\n");
			//printf(".");
			continue;
		}


		/*
		 * If we get here, we passed all the criteria.  We can stop
		 * looping and generating new exponents and running through tests
		 */
		break;
	}
	while(1);

	if(modInvRes != NULL)
		free(modInvRes);

	printf("done\n");

	return true;
}


/*
 * Generate inverse of Exponent
 *
 */
bool genNumModInv(uint32_t *num, uint32_t *mod, uint32_t szWds, uint32_t *res){

	int i;
	uint32_t szNum = szWds, szMod = szWds;

	/* update size if MSWs are 0's */
	while(szNum>1 && num[szNum-1]==0)
		szNum--;

	while(szMod>1 && mod[szMod-1]==0)
		szMod--;

	if(getModInv(num, szNum, mod, szMod, res) != true)
		return false;
	else{
		return true;
	}
}



/*
 * genMsg() - Generate a message.
 *
 * Use the random number generator to generate a random number as a message.  The message, m,
 * needs to be 0 <= m < n where n is the modulo.
 */
bool genMsg(uint32_t *msg, uint32_t *mod, uint32_t szBits){

	int cmpRes = 0, invRes, i;
	uint32_t szWds = (szBits+31)/32;


	do{
		if(genRandNum(msg, szBits) != true)
			return false;

		cmpRes = pka_cmp(msg, mod, szWds);
		if(cmpRes != -1)                     // message must be less than the modulus
			continue;

		invRes = pka_modinv(msg, szWds, mod, szWds, NULL);
		if(invRes == 7 || invRes == 31)      // use the PKA modular inverse function to find if GCD(msg, MOD)=1
			continue;
		else
			if(invRes == 0)                  // if an inverse if found, GCD(msg, MOD) is 1
				break;
	}while(1);

#ifdef DEBUG
	printf("Generated message:\n");
	printf("\tmsg: ");	for(i=szWds-1; i>=0; i--) printf("%08x", msg[i]);
	printf("\n\tmod: "); for(i=szWds-1; i>=0; i--)	printf("%08x", mod[i]); printf("\n");
#endif

	return true;

}


/*
 * rsaEncrypt() - Encrypt the message
 *
 */
bool rsaEncrypt(keypair_t *kp, uint32_t *msg_buffer, uint32_t *ct, uint32_t szWds){

	int i;
	uint32_t *res;

	/* RSA encryption is a modular exponentiation using public key */
	res = pka_modexp(kp->pub_exp, szWds, kp->n, msg_buffer, szWds, ct);

	if(res == ct){
#ifdef DEBUG
		printf("encrypted ciphertext:\n\t"); for(i=szWds-1; i>=0; i--)	printf("%08x", ct[i]);	printf("\n");
#endif
		return true;
	}
	else
		return false;

}


/*
 * rsaDecrypt() - Decrypt the message
 *
 */
bool rsaDecrypt(keypair_t *kp, uint32_t *ct, uint32_t *pt, uint32_t szWds, enum decmode md){

	int i;
	uint32_t *res;

	if(md == pair){
		/* RSA decryption is a modular exponentiation using private key */
		res = pka_modexp(kp->priv_exp, szWds, kp->n, ct, szWds, pt);
	}
	else{
		res = pka_modexpcrt(kp->dp, kp->dq, kp->p, kp->q, kp->qinv, ct, szWds, pt) ;
	}

	if(res == pt){
#ifdef DEBUG
		printf("decrypted plaintext:\n\t"); for(i=szWds-1; i>=0; i--)	printf("%08x", pt[i]);printf("\n");
#endif
			return true;
	}
	else
		return false;


}

/*
 * testDecryptedOutput() - compare buffers
 */
void testDecryptedOutput(uint32_t *buf1,  uint32_t *buf2, uint32_t wds){
	if(pka_cmp(buf1, buf2, wds)==0)
		printf("...PASS!\n");
	else
		printf("...FAIL!\n");

}

