#include <sys/platform.h>
#include <stdlib.h>
#include <stdio.h>

#include "math_utilities.h"
#include "pka_functions.h"
#include "RSAexample.h"

/*
 * isPrime()
 *
 * Tests for primality using a simple test.  Not the most efficient.
 */
bool isPrime(uint32_t *num_buffer, uint32_t szWds){

	uint32_t *ii, *i, *i2, *res, *rem;
	int cmpRes, t, j;
	uint32_t num6 = 6, num2 = 2,  realSz, tmp, ret;
	bool prime=false;

	uint32_t curState = 0;
	const uint32_t transitionMatrix[3][2] = {{0, 1}, {2, 0}, {1, 2}};

	if(szWds == 0){
		printf("isPrime() Error: input size 0!\n");
		exit(1);
	}

	// adjust szWds if needed
	while(num_buffer[szWds-1]==0)
		szWds--;

#if defined(DEBUG) && DEBUG >=3
	printf("checking for prime: \n\t");
		for(t=szWds-1; t>=0; t--) printf("%08x", num_buffer[t]); printf("\n");
#endif

	if(szWds==1)
		if(num_buffer[0] <= 1)
			return false;
		else
			if(num_buffer[0] <= 3)
				return true;

	if((num_buffer[0] & 1) == 0){     // large even number can't be prime
		num_buffer[0] |= 1;          // make odd and test before going back and generating another number.
#if defined(DEBUG) && DEBUG >=3
		printf("isPrime(): Setting LSB.  Input number is even.  Obviously not prime\n");
#endif
	}

	// n mod 3 = 0?
	for(t=0; t<szWds; t++){
		tmp = num_buffer[t];
		for(j=0; j<32; j++){
			curState = transitionMatrix[curState][(tmp&1)];
			tmp >>= 1;
		}
	}
	if(curState == 0)                // not prime if the number is >3 and divisible by 3
		return false;


	ii  = malloc(szWds * sizeof(uint32_t) * 2);
	i   = malloc(szWds * sizeof(uint32_t) * 2);
	i2  = malloc(szWds * sizeof(uint32_t) * 2);
	rem = malloc(szWds * sizeof(uint32_t) * 2);


	if(ii == NULL || i == NULL || i2 == NULL){
		printf("Can't allocate memory for primality testing\n");
		exit(1);
	}

	i[0] = 5;
	res  = pka_mult(i, szWds, i, szWds, ii);
#if defined(DEBUG) && DEBUG >=3
	if(res != ii)
		printf("isPrime() Error: pka_mult failure\n");
#endif

	cmpRes = pka_cmp(ii, num_buffer,szWds);
#if defined(DEBUG) && DEBUG >=3
	if(cmpRes == -2)
		printf("isPrime() Error: pka_cmp failure\n");
#endif

	while(cmpRes == 0 || cmpRes == -1 ){                          // while( i*i <= n )

		res = pka_add(i, szWds, &num2, 1, i2);                    // i+2
#if defined(DEBUG) && DEBUG >=3
		if(res != i2)
			printf("isPrime() Error: pka_add failure\n");
#endif

		realSz = szWds;
		while(i[realSz-1] == 0)
			realSz --;

		if(realSz == 1){                                          // subtract until we find a remainder.  To use PKA, divisor needs to be >32 bits
			ret = aModB(num_buffer, szWds, i[0]);

			if(ret==0){
#if defined(DEBUG) && DEBUG >=3
				printf("n mod i = 0 ... fail\n");
#endif
				prime = false;
				goto isPrime_end;
			}
		}else{
			res = pka_div(num_buffer, szWds, i, realSz, NULL, rem);   // n mod i
#if defined(DEBUG) && DEBUG >=3
			if(res != NULL)
				printf("isPrime() Error: pka_div failure\n");
#endif
			for(t=0, tmp=0; t<realSz; t++)                         // if result = 0, then not prime
				tmp |= rem[t];
			if(tmp == 0){
#if defined(DEBUG) && DEBUG >=3
				printf("n mod i = 0 ... fail\n");
#endif
				prime = false;
				goto isPrime_end;
			}
		}

		realSz = szWds;
		while(i2[realSz-1] == 0)
			realSz --;
		if(realSz == 1){                                          // subtract until we find a remainder.  To use PKA, divisor needs to be >32 bits
			ret = aModB(num_buffer, szWds, i2[0]);
			if(ret==0){
#if defined(DEBUG) && DEBUG >=3
				printf("n mod i+2 = 0 ... fail\n");
#endif
				prime = false;
				goto isPrime_end;
			}
		}else{
			res = pka_div(num_buffer, szWds, i2, realSz, NULL, rem);  // n mod (i+2)
#if defined(DEBUG) && DEBUG >=3
			if(res != NULL)
				printf("isPrime() Error: pka_div failure\n");
#endif
			for(t=0, tmp=0; t<realSz; t++)                            // if result = 0, then not prime
				tmp |= rem[t];
			if(tmp == 0){
#if defined(DEBUG) && DEBUG >=3
				printf("n mod i+2 = 0 ... fail\n");
#endif
				prime = false;
				goto isPrime_end;
			}
		}

		res = pka_add(i, szWds, &num6, 1, i);                      // i+=6
#if defined(DEBUG) && DEBUG >=3
		if(res != i)
			printf("isPrime() Error: pka_add failure\n");
#endif

#if defined(DEBUG) && DEBUG >=3
		printf("isPrime(): i: 0x%x\n", i[0]);
#endif

		res  = pka_mult(i, szWds, i, szWds, ii);
#if defined(DEBUG) && DEBUG >=3
		if(res != ii)
			printf("isPrime() Error: pka_mult failure\n");
#endif

		cmpRes = pka_cmp(ii, num_buffer,szWds);
#if defined(DEBUG) && DEBUG >=3
		if(cmpRes == -2)
			printf("isPrime() Error: pka_cmp failure\n");
#endif
	}

	prime = true;                                                   // if we made it through the loop, then it's prime

isPrime_end:
	if(ii != NULL)
		free(ii);
	if(i != NULL)
		free(i);
	if(i2 != NULL)
		free(i2);
	if(rem != NULL)
		free(rem);

	return prime;


}






/*
 * aModB() - A MOD  B;   A%B
 *
 * This function does division by scaling the divisor and subtracting.
 *
 * Input
 * 	a - pointer the large number buffer
 * 	szWds - size of vector a in 32-bit words
 * 	b - divisor (max is 32bits)  (if B is greater than 1 words, use PKA engine)
 *
 */
uint32_t aModB(uint32_t *a, uint32_t szWds, uint32_t b){

	int i;
	uint32_t *tmpBuf, x, *res, *xBuf, rem;

	/* allocate memory for temporary working buffer */
	tmpBuf = malloc(szWds * sizeof(uint32_t));
	if(tmpBuf == NULL){
		printf("aModB() Error: Can't allocate memory for temp buffer\n");
		exit(1);
	}

	/* allocate another buffer to use with PKA subtraction */
	xBuf = malloc(szWds * sizeof(uint32_t));
	if(xBuf == NULL){
		printf("aModB() Error: Can't allocate memory for xBuf for pka_sub\n");
		exit(1);
	}

	/* copy dividend to temporary working buffer */
	for(i=0; i<szWds; i++)
		tmpBuf[i] = a[i];


	while(szWds >= 1){

		if(tmpBuf[szWds-1] >= b){
			for(x=b; ((x<<1) <= tmpBuf[szWds-1]) && (((x<<1) & 0x80000000) == 0)  ; x <<= 1)
				;
			tmpBuf[szWds-1] -= x;
		}

		/* The final result in the LSW is either 0 or a remainder less than b */
		if(szWds==1 && tmpBuf[0] < b)
			break;

		/* prepare xBuf with the maximum  b * 2^n value where xBuf is 1 word less than tmp_buf */
		if(szWds>1 && tmpBuf[szWds-1]!=0 && tmpBuf[szWds-1]<b){
			for(xBuf[szWds-2]=b; (xBuf[szWds-2] & 0x80000000) == 0; xBuf[szWds-2] <<= 1)
				;
		}

		/* use PKA to subtract xBuf from tmp_buf until the MSW of tmp_buf is 0 and we can work within a word boundary again */
		while(szWds>1 && tmpBuf[szWds-1]!=0 && tmpBuf[szWds-1]<b){
			res = pka_sub(tmpBuf, szWds, xBuf, szWds-1, tmpBuf);
			if(res != tmpBuf){
				printf("aModB() Error.  Error from pka_sub()\n");
				exit(1);
			}
		}

		/* decrement the size of tmpBuf if the MSW is 0 */
		if(tmpBuf[szWds-1]==0)
			szWds--;
	}

	rem = tmpBuf[0];
	free(tmpBuf);
	free(xBuf);

	return rem;

}


/*
 * getModInv() - Calculates the modular inverse of a large number using PKA engine
 *
 * Since the modular inverse operation using the PKA is only to a modulus being odd,
 * and since φ(n) is even since p and q are odd primes, the modular inverse of the exponent
 * is calculated with d = ModInv(e, φ) = (1 + (φ * (e - ModInv(φ, e)))/e
 */

bool getModInv(uint32_t *num, uint32_t numSzWds, uint32_t *modulo, uint32_t moduloSzWds, uint32_t *modInvRes){
	uint32_t pkaModInvStat;
	uint32_t *res, *tmpOutput;
	uint32_t num1 = 1;
	int i;


	res = NULL;
	tmpOutput = NULL;


	/* input is odd.  We can directly use the PKA engine */
	if((modulo[0] & 1) != 0){
		pkaModInvStat = pka_modinv(num, numSzWds, modulo, moduloSzWds, modInvRes);
		if(pkaModInvStat != 0){
			printf("Error: Could not calculate inverse of exponent: return code: %d\n", pkaModInvStat);
			return false;
		}
	}
	/* The number to find the inverse is even.  Use alternate method */
	else{
		/* step 1: ModInv(φ, e) */
		pkaModInvStat = pka_modinv(modulo, moduloSzWds, num, numSzWds, modInvRes);
		if(pkaModInvStat != 0){
			printf("Error: Could not calculate inverse of exponent: return code: %d\n", pkaModInvStat);
			return false;
		}
		/*
		 * step 2: subtract result from exp:  e - ModInv(φ, e)
		 * Use modInvRes to store this temp result
		 */
		res = pka_sub(num, numSzWds, modInvRes, numSzWds, modInvRes);
		if(res != modInvRes){
			printf("Error: getModInv(): Error from pka_sub()\n");
			exit(1);
		}

		/* we need an output buffer for the result of the multiply that is twice the size of the input */
		tmpOutput = malloc((numSzWds+moduloSzWds) * 4 );
		if(tmpOutput == NULL){
			printf("Error: getModInv(): Could not allocate memory for output of multiply\n");
			exit(1);
		}

		/* step 3: multiply result with φ  */
		res = pka_mult(modulo, moduloSzWds, modInvRes, numSzWds, tmpOutput);
		if(res != tmpOutput){
			printf("Error: getModInv(): Error from pka_mult()\n");
			exit(1);
		}

		/* step 4: add 1 */
		res = pka_add(&num1, 1, tmpOutput, numSzWds * 2, tmpOutput);
		if(res != tmpOutput){
			printf("Error: getModInv(): Error from pka_add()\n");
			exit(1);
		}

		/* step 5: divide by exponent */
		res = pka_div(tmpOutput, numSzWds * 2, num, numSzWds, modInvRes, NULL);
		if(res != modInvRes){
			printf("Error: getModInv(): Error from pka_div()\n");
			exit(1);
		}
	}


#if defined(DEBUG) && DEBUG >=2
		printf("getModInv():\n\tnum to find inverse:\n\t");
		for(i=numSzWds-1; i>=0; i--) printf("%08x", num[i]); printf("\n");

		printf("\tmodulus:\n\t");
		for(i=moduloSzWds-1; i>=0; i--)	printf("%08x", modulo[i]); printf("\n");

		printf("\tInverse found:\n\t");
		for(i=numSzWds-1; i>=0; i--) printf("%08x", modInvRes[i]); printf("\n");

#endif

		/* verify that e*d=1 MOD phi */
		if(tmpOutput == NULL)
			tmpOutput = malloc(numSzWds * 4 *2);
		if(tmpOutput == NULL){
			printf("Error: Can't malloc for debug\n");
			exit(1);
		}
		res = pka_mult(num, numSzWds, modInvRes, numSzWds, tmpOutput);
		if(res != tmpOutput){
			printf("Error: Problem with pka_mult in debug\n");
			exit(1);
		}

		if(moduloSzWds==1){
			tmpOutput[0] = aModB(tmpOutput, numSzWds, modulo[0]);
		}
		else{
			res = pka_div(tmpOutput, numSzWds * 2, modulo, moduloSzWds, NULL, tmpOutput);
			if(res != NULL){   // return should be null since our output division result ptr is NULL
				printf("Error: Problem with pka_div in debug\n");
				exit(1);
			}
		}
#if defined (DEBUG) && DEBUG >=2
		printf("Verify: num * num_inv = %d\n", tmpOutput[0]);
#endif

		if(tmpOutput != NULL)
			free(tmpOutput);




	return true;
}


